/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/

#ifndef _common_h
#define _common_h
#include "xdp_jpeg_hdr.h"

#if defined(_MSC_VER)
#define _ALIGNED
#define COPYSIGN _copysign
#else
#define COPYSIGN copysign
#endif

#define XT_ERROR_OK 1
#define XT_ERROR_NULL_POINTER -1
#define XT_ERROR_ALLOC  -2
#define XT_ERROR_INVALID_PROFILE  -3
#define XT_ERROR_CHECKSUM   -4
#define XT_ERROR_INVALID_INFO -5

#define XDP_SAFE_DELETE( p ) { if( p ) { free ( p ); ( p ) = NULL; }};

#define XDP_SWAP_SHORT16(a,b) \
      a=((((unsigned int)((b) & 0x00FF)) << 8) | \
         (((unsigned int)((b) & 0xFF00)) >> 8));

#define XDP_SWAP_SHORT32(a,b) \
         a=((((unsigned int)((b) & 0x000000FF)) << 24) | \
         (((unsigned int)((b) & 0x0000FF00)) << 8) | \
         (((unsigned int)((b) & 0x00FF0000)) >> 8) | \
         (((unsigned int)((b) & 0xFF000000)) >> 24));
#define BIT_SET(a,b) ((a) |= (1<<(b)))
#define BIT_CLEAR(a,b) ((a) &= ~(1<<(b)))
#define BIT_CHECK(a,b) ((a) & (1<<b))

// Default RGB to XYZ conversion matrix
static float g_RGB2XYZ[3][3]={{0.412453f, 0.357580f,  0.180423f},
                     {0.212671f, 0.715160f,  0.072169f},
                     {0.019334f, 0.119193f,  0.950227f}};

// Default XYZ to RGB conversion matrix
static float g_XYZ2RGB[3][3]={{3.240479f,-1.537150f, -0.498535f},
                     {-0.969256f,  1.875992f,  0.041556f},
                     {0.055648f, -0.204043f,  1.057311f}};

#define DEFAULT_P1 0.04045f
#define DEFAULT_P2 2.4f
#define DEFAULT_P3 0.055f

unsigned int getStartMillisecs(const char *message);
unsigned int getEndMillisecs();
#define START_TIMER(a) getStartMillisecs(a);
#define END_TIMER() getEndMillisecs();

enum {
     XT_P1 = 0,
     XT_P2,
     XT_P3,
     XT_P4
};
void compile_box(JPXT_SUPERBOX *super_box,unsigned int payload_size,unsigned int TBox);
float swap_float(float value);
float toFloat(unsigned char in_byte,int Rw, int e);
unsigned char toByte(float in_value, int Rw, int e);
#endif
