/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "image.h"
#include "encoder.h"

#include <string.h>
//#include <stdint.h>
#include "common.h"
#include "hdr_io.h"
#include "xdp_jpeg_hdr.h"
#include "checksum.h"
#include "tone_mapping.h"

extern int g_clip_input;
extern int g_Rw;
extern int g_e;
extern int g_tmo;
extern int g_jpeg_dct;
extern int g_huffman_opt;
extern int g_lossless;
extern int g_save_resi_bitstream;

extern float g_ocon_p1;
extern float g_ocon_p2;
extern float g_ocon_p3;
extern float g_ocon_p4;
extern int g_use_custom_ocon;

#ifndef max
#define max( a, b ) ( ((a) > (b)) ? (a) : (b) )
#endif

#ifndef min
#define min( a, b ) ( ((a) < (b)) ? (a) : (b) )
#endif

// Autobalance image at the right exposure
// This function scale values of the input image.
int xdp_HDRAutobalance_hdr(float *hdr_image, int width, int height,int channels,float scale_factor,float *out_value)
{
    unsigned int c;
    float averagef=0.0f;

	unsigned int img_size;
	float factor;

    unsigned int numpix=width*height*channels;
    for (c=0;c<numpix;c+=channels)
    {
		float m;
        float r=(float)hdr_image[c];
        if (channels==3)
        {
            float g=(float)hdr_image[c+1];
            float b=(float)hdr_image[c+2];
            m=(r+g+b)/3.0f;
        }
        else
            m=r;
        averagef=averagef+m;
        
    }
    
    img_size=width*height;

    averagef=averagef / img_size;

    factor = scale_factor/averagef;

    *out_value=1.0f/factor;

    for (c=0;c<((unsigned int)width*(unsigned int)height*channels);c+=channels)
    {
        hdr_image[c]=hdr_image[c]*factor;
        if (channels==3)
        {
            hdr_image[c+1]=hdr_image[c+1]*factor;
            hdr_image[c+2]=hdr_image[c+2]*factor;
        }
    }
    return 1;
}

// Autobalance version 2 image at the right exposure
// This function scale values of the input image.
int xdp_HDRAutobalance_hdr2(float *hdr_image, int width, int height,int channels,float scale_factor,float *out_value, float tmo[256],FILE_IMAGE *ppm8)
{
    unsigned int c;
    float averagef=0.0f;

    unsigned int img_size;
    float factor;
	float max;
	unsigned int offset=0;
	float clip;
	unsigned char *p_ppm8;
    unsigned int numpix=width*height*channels;
    for (c=0;c<numpix;c+=channels)
    {
        float m;
        float r=(float)hdr_image[c];
        if (channels==3)
        {
            float g=(float)hdr_image[c+1];
            float b=(float)hdr_image[c+2];
            m=(r+g+b)/3.0f;
        }
        else
            m=r;
        averagef=averagef+m;

    }

    img_size=width*height;

    averagef=averagef / img_size;

    factor = scale_factor/averagef;

    *out_value=1.0f/factor;

    max=1.0f;
    

    //clip = ((1.0f/factor) * tmo[1]) / (1.0f+1e-6);
    clip = 1e-6;

    p_ppm8=ppm8->data;

    for (c=0;c<((unsigned int)width*(unsigned int)height);c+=1)
    {
        int f;
        for (f=0;f<channels;f++)
        {
            double hdrv = (hdr_image[offset+f])*factor;
            double ldrv = tmo[p_ppm8[offset+f]];
            if (hdrv > clip) {
                float quot = ldrv / hdrv;
                if (quot>max)
                {
                    max=quot;
                }
            }
        }
        offset+=channels;
    }
    if (max > 1.0) {
        factor *= max;
        *out_value =1.0f/factor;
    }

    for (c=0;c<((unsigned int)width*(unsigned int)height*channels);c+=channels)
    {
        hdr_image[c]=hdr_image[c]*factor;
        if (channels==3)
        {
            hdr_image[c+1]=hdr_image[c+1]*factor;
            hdr_image[c+2]=hdr_image[c+2]*factor;
        }
    }

    return 1;
}


float dynamic_range(float *hdr_image, int width, int height,int channels)
{
    unsigned int c;
    float mx=0;
    float mn=65535.f;
    float dr;

    unsigned int numpix=width*height*channels;
    for ( c=0;c<numpix;c+=channels)
    {
        float g,b;
        float r=(float)hdr_image[c];
        mn=min(mn,r);
        mx=max(mx,r);
        if (channels==3)
        {
            g=(float)hdr_image[c+1];
            b=(float)hdr_image[c+2];

            mn=min(mn,g);
            mn=min(mn,b);

            mx=max(mx,g);
            mx=max(mx,b);
        }
    }
    mn+=1e-6;
    dr = (float) log10(mx/mn);
    if(dr != dr) { dr = 1.0; }
    printf("Dynamic range : %f\n",dr);
    return dr;
}
// Calculate gamma for the HDR values, 1st pass
float xdp_HDRAutoGamma(float *hdr_image, int width, int height,int channels)
{
    
	float min_val;
    float minrgb = 0.5;
    unsigned long c=1;
    min_val=1.0;

    for (c=0;c<((unsigned int)width*(unsigned int)height)*channels;c++)
	{
		float curval;
		if (hdr_image[c]>0.00001)
			curval=(float)1.0/hdr_image[c];
		else
			continue;
		if (curval>0.0)
		{
			if (curval<min_val)
				min_val=curval;
		}
	}

    return min_val < minrgb?log(min_val)/log(minrgb):1.0f;
}

FILE_IMAGE *xdp_pre_encoding(FILE_IMAGE *in_image, float lg,int srgb)
{
	FILE_IMAGE*tm24;
    FILE_IMAGE *tmimage=image_convert_float(in_image);
    if (lg!=1.0f && srgb==0)
        image_linearize(tmimage,lg);

    // Linearize tone mapped image with inverse srgb
    if (srgb!=0 && lg==1.0f)
        srgb_to_linear(tmimage);

    shift_ldr_image(tmimage);
    // Apply again gamma on float tone mapped image
    if (lg!=1.0f && srgb==0)
        image_linearize(tmimage,1.0f/lg);

    // Apply again srgb on float tone mapped image
    if (srgb!=0 && lg==1.0f)
        linear_to_srgb(tmimage);
    tm24=image_convert_24(tmimage);

    return tm24;
}


// Encode HDR image in XDepth images
int xdp_HDREncode(FILE_IMAGE *hdr_image,int pre_encoded,MEMORY_FILE_IMAGE **out_24bit_image_prencoded,FILE_IMAGE **out_24bit_image, FILE_IMAGE **out_hdr_image, float gamma_hdr,int auto_hdr_gamma,float *generated_auto_gamma,int auto_balance,
                  float *out_exp_value,float scale_factor,float lg, FILE_IMAGE *input_tone_mapped_image,float *min_r,float *min_g,float *min_b,float *max_r,float *max_g,float *max_b,
                  int input_xyz,int srgb,float base_matrix[], float resi_matrix[],int resi_identity,JPG_INFO *jpg_info)
{
    float *hdr_float_buffer = (float *) hdr_image->data;
    float *hdr_rgb_float_buffer=NULL;
    FILE_IMAGE *rgb_hdr_image=NULL;
    FILE_IMAGE *cur_tone_mapped=NULL;
    unsigned width=hdr_image->width;
    unsigned height=hdr_image->height;
    unsigned long image_size;
    unsigned line_8bit;
    unsigned line_96bit;
    unsigned int c;
    float gamma_out;
    float *out_hdr=NULL;
    FILE_IMAGE *ldr_24bit;
    FILE_IMAGE *hdr_24bit;
    unsigned y,x;
    unsigned char *tmp_hdr;
    unsigned char *lbitshdr8;
    float out_exposure_value;
    float mnr=HUGE_VAL,mng=HUGE_VAL,mnb=HUGE_VAL;
    float mxr=-HUGE_VAL,mxg=-HUGE_VAL,mxb=-HUGE_VAL;
    unsigned char *tmp_reinhard=NULL;
    float *tmimage;
    FILE_IMAGE *reinhard;
    FILE_IMAGE *tmo_image;
	unsigned char *input_image;
	unsigned char *out_image;
	unsigned char *tone_mapped;
	unsigned stride;
	unsigned int h=0;
    int i;
	float tonemapping[256];
    unsigned input_channels=3;
	FILE_IMAGE *residual;
	unsigned char *i_tonemapped;
	unsigned istride;
    if (hdr_image->grayscale==1)
        input_channels=1;

    image_size=width*height*input_channels;
    line_8bit=width * input_channels;
    line_96bit=(width * input_channels)*sizeof(float);
    out_hdr=(float *)malloc(image_size*sizeof(float));
	
    if (out_hdr==NULL)
        return 0;

    if (input_xyz==1)
    {
        if (resi_identity==0)
            rgb_hdr_image=image_convert_xyz_rgb(hdr_image,base_matrix);
        else
            rgb_hdr_image=hdr_image;
    }

    // Allocate space for the output ldr image
    ldr_24bit = allocate_image( width, height, 24,input_channels);
    if (ldr_24bit == NULL)
    {
        free(out_hdr);
        return 0;
    }

    // Allocate space for the output hdr image
    hdr_24bit = allocate_image( width, height, 24,input_channels);
    if (hdr_24bit == NULL)
    {
        free(out_hdr);
        free_image (ldr_24bit);
        return 0;
    }
    if (input_channels==1)
    {
        ldr_24bit->grayscale=1;
        hdr_24bit->grayscale=1;
    }
    
    // Generate LUT for encoding
    for (i=0;i<256;i++)
    {
        float x=i/255.0f;
        if (srgb!=0)
        {
            if (x < 0.04045)
                tonemapping[i] = (x / 12.92f);
            else
                tonemapping[i] = (powf((x + 0.055f)/(1.055f),2.4f));
        }
        else
            tonemapping[i] = powf(x,lg);
    }

    cur_tone_mapped=input_tone_mapped_image;
    if (input_tone_mapped_image==NULL)
    {
        FILE_IMAGE *hdr_copy=image_copy(hdr_image);
        float *p=(float*)hdr_copy->data;
        float autoexp_range=dynamic_range(p,width,height,input_channels);
        xdp_HDRAutobalance_hdr(p,width,height,input_channels, autoexp_range,&out_exposure_value);
        if (input_xyz==0)
        {
			FILE_IMAGE*reinhard24;
			MEMORY_FILE_IMAGE *out_mem_jpeg;
            reinhard=tone_mapping(g_tmo,hdr_copy,lg);
            reinhard24=image_convert_24(reinhard);

            if (pre_encoded==1)
            {
                out_mem_jpeg=jpeg_write_mem(reinhard24,jpg_info->compression_level,JPEG_SUBSAMPLE_411,jpg_info->dct_type,jpg_info->optimized_huffman,0);
                *out_24bit_image_prencoded=out_mem_jpeg;
                free_image(reinhard);
                free_image(reinhard24);
                reinhard24=jpeg_load_mem(out_mem_jpeg);
                cur_tone_mapped=reinhard24;
                reinhard=image_convert_float(reinhard24);

            }
            else
            {
                cur_tone_mapped=reinhard24;
                *out_24bit_image=image_copy(reinhard24);
            }
            free_image(hdr_copy);
        }
        else
        {
			FILE_IMAGE*reinhard24;
			MEMORY_FILE_IMAGE *out_mem_jpeg;
            FILE_IMAGE *rgb_copy=image_copy(rgb_hdr_image);
            float autoexp_range=dynamic_range(hdr_float_buffer,width,height,input_channels);
            xdp_HDRAutobalance_hdr((float*)rgb_copy->data,width,height,input_channels, autoexp_range,&out_exposure_value);
            reinhard=tone_mapping(g_tmo,rgb_copy,lg);
            reinhard24=image_convert_24(reinhard);
            if (pre_encoded==1)
            {
                out_mem_jpeg=jpeg_write_mem(reinhard24,jpg_info->compression_level,JPEG_SUBSAMPLE_411,jpg_info->dct_type,jpg_info->optimized_huffman,0);
                *out_24bit_image_prencoded=out_mem_jpeg;
                free_image(reinhard);
                free_image(reinhard24);
                reinhard24=jpeg_load_mem(out_mem_jpeg);
                cur_tone_mapped=reinhard24;
                reinhard=image_convert_float(reinhard24);
            }
            else
            {
                cur_tone_mapped=reinhard24;
                *out_24bit_image=image_copy(reinhard24);
            }
            free_image(rgb_copy);
        }
    }

    // Autobalance image, the output float buffer is now gamma 2.2f (or gamma ldr)
    if (auto_balance)
    {
        float autoexp_range=dynamic_range(hdr_float_buffer,width,height,input_channels);
        if (pre_encoded==1)
            xdp_HDRAutobalance_hdr(hdr_float_buffer,width,height,input_channels, autoexp_range,&out_exposure_value);
        else
            xdp_HDRAutobalance_hdr2(hdr_float_buffer,width,height,input_channels, autoexp_range,&out_exposure_value,tonemapping,cur_tone_mapped);
        *out_exp_value=out_exposure_value;
        printf("Autoexposure value : %f\n",out_exposure_value);
    }
    if (auto_hdr_gamma)
    {
        if (input_xyz==0)
        {
            // Calculate gamma value for the hdr data
            gamma_out=xdp_HDRAutoGamma(hdr_float_buffer,width,height,input_channels);
        }
        else
        {
            hdr_rgb_float_buffer=(float *)rgb_hdr_image->data;
            // Calculate gamma value for the hdr data
            gamma_out=xdp_HDRAutoGamma(hdr_rgb_float_buffer,width,height,input_channels);
        }

        if (gamma_out>8.0)
            gamma_out=8.0;


        gamma_hdr=gamma_out;
    }
    printf("LDR gamma: %f\n",lg);
    printf("HDR gamma : %f\n",gamma_hdr);

    *generated_auto_gamma=gamma_hdr;
    *min_r=HUGE_VAL;
    *min_g=HUGE_VAL;
    *min_b=HUGE_VAL;

    *max_r=-HUGE_VAL;
    *max_g=-HUGE_VAL;
    *max_b=-HUGE_VAL;


    // Scale 8-bit tone mapped image in float
    if (input_tone_mapped_image!=NULL)
    {
        if (pre_encoded==1)
        {
            MEMORY_FILE_IMAGE *out_mem_jpeg;
            FILE_IMAGE *out=input_tone_mapped_image;

            out_mem_jpeg=jpeg_write_mem(out,jpg_info->compression_level,JPEG_SUBSAMPLE_411,jpg_info->dct_type,jpg_info->optimized_huffman,0);
            *out_24bit_image_prencoded=out_mem_jpeg;
            free_image(out);
            input_tone_mapped_image=jpeg_load_mem(out_mem_jpeg);
            cur_tone_mapped=input_tone_mapped_image;
            reinhard=image_convert_float(input_tone_mapped_image);
        }
        else
        {
            FILE_IMAGE *out=image_copy(input_tone_mapped_image);
            *out_24bit_image=out;
            reinhard=image_convert_float(input_tone_mapped_image);
        }
    }

    // Keep a copy for output jpeg-xt file
    tmo_image=image_copy(reinhard);

    // Convert input tone mapped image in XYZ
    // Linearize tone mapped image with inverse gamma
    if (lg!=1.0f && srgb==0)
        image_linearize(reinhard,lg);

    // Linearize tone mapped image with inverse srgb
    if (srgb!=0 && lg==1.0f)
    {
        srgb_to_linear(reinhard);
    }


    if (input_xyz==1)
    {
        float matrix[]={g_RGB2XYZ[0][0],g_RGB2XYZ[0][1],g_RGB2XYZ[0][2],
                      g_RGB2XYZ[1][0],g_RGB2XYZ[1][1],g_RGB2XYZ[1][2],
                      g_RGB2XYZ[2][0],g_RGB2XYZ[2][1],g_RGB2XYZ[2][2]};
        FILE_IMAGE *xyz_image=image_convert_rgb_xyz(reinhard,matrix);
        free_image(reinhard);
        reinhard=xyz_image;
    }

    tmimage=(float *)reinhard->data;

    input_image=(unsigned char*)hdr_float_buffer;
    out_image=(unsigned char*)out_hdr;
    tone_mapped=reinhard->data;
    stride=(width*input_channels)*sizeof(float);
    for (h=0;h<height;h++)
    {
        unsigned int hx=0;
        float *hdr_image=(float*)input_image;
        float *ptm_image=(float*)tone_mapped;
        float *pout=(float*)out_image;
        c=0;
        for (hx=0;hx<width;hx++)
        {

                pout[c]=1.0f;
                if (hdr_image[c]>HDR_MIN_VAL)
                    pout[c]=(ptm_image[c])/hdr_image[c];

                pout[c]=ptm_image[c]==0?1.0f:pout[c];

                if (input_channels==3)
                {
                    pout[c+1]=1.0f;
                    pout[c+2]=1.0f;



                    if (hdr_image[c+1]>HDR_MIN_VAL)
                        pout[c+1]=(ptm_image[c+1])/hdr_image[c+1];

                    if (hdr_image[c+2]>HDR_MIN_VAL)
                        pout[c+2]=ptm_image[c+2]/hdr_image[c+2];



                    pout[c+1]=ptm_image[c+1]==0?1.0f:pout[c+1];
                    pout[c+2]=ptm_image[c+2]==0?1.0f:pout[c+2];
                }
                if (input_xyz==0)
                {
                    mnr=pout[c]<mnr?pout[c]:mnr;
                    mxr=pout[c]>mxr?pout[c]:mxr;
                    if (input_channels==3)
                    {
                    mng=pout[c+1]<mng?pout[c+1]:mng;
                    mnb=pout[c+2]<mnb?pout[c+2]:mnb;

                    mxg=pout[c+1]>mxg?pout[c+1]:mxg;
                    mxb=pout[c+2]>mxb?pout[c+2]:mxb;
                    }
                }


            c+=input_channels;
        }
        input_image+=stride;
        out_image+=stride;
        tone_mapped+=stride;
    }

    residual=allocate_image(width,height,96,input_channels);
    residual->data=(unsigned char *)out_hdr;
    residual->data_size=image_size*sizeof(float);
    if (input_xyz==0)
    {
            *min_r=mnr;
            *max_r=mxr;
        if (input_channels==3)
        {
            *min_g=mng;
            *min_b=mnb;

            *max_g=mxg;
            *max_b=mxb;
        }
    }
    else
    {
        FILE_IMAGE *residual_RGB=image_convert_xyz_rgb(residual,resi_matrix);
        image_minmax_values(residual_RGB,min_r,min_g,min_b,max_r,max_g,max_b);
        free(residual);
    }
/*
    *max_r=*max_r>1.0f?1.0f:*max_r;
    *max_g=*max_g>1.0f?1.0f:*max_g;
    *max_b=*max_b>1.0f?1.0f:*max_b;
*/
/*
    // Apply again gamma on float tone mapped image
    if (lg!=1.0f && srgb==0)
        image_linearize(reinhard,1.0f/lg);

    // Apply again srgb on float tone mapped image
    if (srgb!=0 && lg==1.0f)
        linear_to_srgb(reinhard);
*/

    if (input_xyz==0)
    {
        tmp_reinhard=reinhard->data;
        free_image(tmo_image);
    }
    else
    {
        tmp_reinhard=tmo_image->data;
        free_image(reinhard);
    }

    if (input_channels==3)
    {
        printf("max values : %f, %f, %f\n",*max_r,*max_g,*max_b);
        printf("min values : %f, %f, %f\n",*min_r,*min_g,*min_b);
    }
    else
    {
        printf("max values : %f\n",*max_r);
        printf("min values : %f\n",*min_r);
    }
    // Store LDR tone mapped image, and HDR fractional part in two 24 bit images
	tmp_hdr=(unsigned char *)out_hdr;
    lbitshdr8 = hdr_24bit->data;
    for( y=0; y<height; y++)
	{
		float *fpixelhdr;
		unsigned char *pixelhdr8;
		int off=0;

		fpixelhdr=(float *)tmp_hdr;
		pixelhdr8 = (unsigned char*)lbitshdr8;
		
		for( x=0; x<width; x++)
		{

            if (input_xyz==1)
            {

                float X,Y,Z;
                float R,G,B;

                X=fpixelhdr[off];
                Y=fpixelhdr[off+1];
                Z=fpixelhdr[off+2];

                if (resi_identity==0)
                {
                    R = (resi_matrix[0] * X ) + (resi_matrix[1] * Y ) + (resi_matrix[2] * Z);
                    G = (resi_matrix[3] * X ) + (resi_matrix[4] * Y ) + (resi_matrix[5] * Z);
                    B = (resi_matrix[6] * X ) + (resi_matrix[7] * Y ) + (resi_matrix[8] * Z);
                }
                else
                {
                    R = X;
                    G = Y;
                    B = Z;
                }

                R = (R-*min_r)/(*max_r-*min_r);
                G = (G-*min_g)/(*max_g-*min_g);
                B = (B-*min_b)/(*max_b-*min_b);

                pixelhdr8[off] = toByte(pow(R,1.0f/gamma_hdr),g_Rw,g_e);
                pixelhdr8[off+1] = toByte(pow(G,1.0f/gamma_hdr),g_Rw,g_e);
                pixelhdr8[off+2] = toByte(pow(B,1.0f/gamma_hdr),g_Rw,g_e);
            }
            else
            {

                    fpixelhdr[off]=(fpixelhdr[off]-mnr)/(mxr-mnr);
                    fpixelhdr[off]=pow(fpixelhdr[off],1.0f/gamma_hdr);
                    pixelhdr8[off] = toByte(fpixelhdr[off],g_Rw,g_e);

                    if (input_channels==3)
                    {
                        fpixelhdr[off+1]=(fpixelhdr[off+1]-mng)/(mxg-mng);
                        fpixelhdr[off+1]=pow(fpixelhdr[off+1],1.0f/gamma_hdr);
                        pixelhdr8[off+1] = toByte(fpixelhdr[off+1],g_Rw,g_e);

                        fpixelhdr[off+2]=(fpixelhdr[off+2]-mnb)/(mxb-mnb);
                        fpixelhdr[off+2]=pow(fpixelhdr[off+2],1.0f/gamma_hdr);
                        pixelhdr8[off+2] = toByte(fpixelhdr[off+2],g_Rw,g_e);
                    }
            }

        off+=input_channels;
       }
        tmp_hdr += line_96bit;
        lbitshdr8 += line_8bit;
	}



    if (input_xyz==0)
        free_image(reinhard);
    else
        free_image(tmo_image);

	free(out_hdr);
    free_image(cur_tone_mapped);

    *out_hdr_image = hdr_24bit;
    return 1;
}

// Encode HDR image in XDEpth-HDR Format
int xdp_encode(FILE_IMAGE *hdr_image,int use_prencode,const char *out_filename,float autobalance, int quality_hdr,int subsamp_hdr,
               int quality_ldr,int subsamp_ldr,float gamma_compatibile,int srgb,float gamma_hdr,int auto_gamma_hdr,
               FILE_IMAGE *input_tone_mapped_img, int input_xyz,float base_matrix[],float resi_matrix[],int resi_identity)
{
    MEMORY_FILE_IMAGE *out_24bit_prencoded=NULL;
    FILE_IMAGE *out_24bit=NULL;
    FILE_IMAGE *out_hdr=NULL;
    MEMORY_FILE_IMAGE *mem_24bit=NULL;
    MEMORY_FILE_IMAGE *mem_hdr_24bit=NULL;
    unsigned width,height;
    int use_autobalance=0;
    float auto_gamma;
    FILE *output_file;
    float out_exposure_value=0;
    unsigned short residualChkSize;
    unsigned short chksize;
    unsigned int off;
    unsigned int w;
    unsigned short chkIdx;
    unsigned int chksum=0;
    JPXT_SUPERBOX  xt_box_header; // Box file type
    JPXT_FTYPE xt_ftype_payload;
    JPXT_SUPERBOX  xt_box_merging; // Merging specification box
    JPXT_CURV  xt_curve[5];
    JPXT_CURV  xt_curve_log[2];
    JPXT_BASE_NON_LINEAR_TRAFO xt_trafo_nlp;
    JPXT_BASE_NON_LINEAR_TRAFO xt_trafo_ir;
    JPXT_BASE_NON_LINEAR_TRAFO xt_trafo_log_base;
    JPXT_BASE_NON_LINEAR_TRAFO xt_trafo_log_resi;
    JPXT_SUPERBOX  xtBoxResidual;
    JPXT_OUT_CONV_BOX xtConversionBox;

    JPXT_FP_LIN_TRAFO xt_lin_trafo[2];
    JPXT_COLOR_TRAFO xt_color_trafo;
    JPXT_RESI_TRAFO xt_resi_trafo;
    float nchunks;
    float min_r,min_g,min_b;
	float max_r,max_g,max_b;
    JPG_INFO jpg_ldr_info;
    int superbox_size;
    int huff=g_huffman_opt;
    int prencoding=use_prencode;
    // Compress input float image in two 24 bit images
    if (autobalance>0)
        use_autobalance=1;

	jpg_ldr_info.compression_level=quality_ldr;
    jpg_ldr_info.dct_type=g_jpeg_dct;
    jpg_ldr_info.lossless=g_lossless;
    jpg_ldr_info.optimized_huffman=g_huffman_opt;
    START_TIMER("Start Encoding")
    if (!xdp_HDREncode(hdr_image,prencoding,&out_24bit_prencoded,&out_24bit,&out_hdr,gamma_hdr,auto_gamma_hdr,&auto_gamma,use_autobalance,&out_exposure_value,autobalance,gamma_compatibile,input_tone_mapped_img,&min_r,&min_g,&min_b,&max_r,&max_g,&max_b,input_xyz,srgb,base_matrix,resi_matrix,resi_identity,&jpg_ldr_info))
        return 0;
    END_TIMER();
    START_TIMER("Compress legacy JPeg");
    if (prencoding==1)
    {
        width = out_24bit_prencoded->width;
        height = out_24bit_prencoded->height;

        mem_24bit = out_24bit_prencoded;
    }
    else
    {
        // Write 24 bit ldr buffer in jpeg format in memory
        mem_24bit = jpeg_write_mem (out_24bit,quality_ldr,subsamp_ldr,g_jpeg_dct,g_huffman_opt,0);
        if (mem_24bit==NULL)
        {
            free_image(out_24bit);
            free_image(out_hdr);
            return 0;
        }

    }

    // Write 24 bit hdr fractional part in jpeg format in memory.
    if (g_lossless==1)
        huff=0;
    mem_hdr_24bit = jpeg_write_mem (out_hdr,quality_hdr,subsamp_hdr,g_jpeg_dct,huff,g_lossless);
    if (mem_hdr_24bit==NULL)
    {
        free_memory_image (mem_24bit);
        free_image(out_24bit);
        free_image(out_hdr);
        return 0;
    }
    END_TIMER();
    START_TIMER("JPEG-XT File format and save");
    if (g_save_resi_bitstream==1)
    {
        FILE *resi_file=fopen("residual.jpg","wb");
        if (resi_file!=NULL)
        {
            fwrite(mem_hdr_24bit->data,1,mem_hdr_24bit->data_size,resi_file);
            fclose(resi_file);
        }
    }
    free_image(out_hdr);

    // Write JPEG XT file here

    // Open file
    output_file=fopen(out_filename,"wb");
    if (output_file==NULL)
    {
        free_memory_image(mem_24bit);
        free_memory_image(mem_hdr_24bit);
        return 0;
    }
    // Calculate checksum on LDR image
    chksum=computeChecksum(mem_24bit);
    fwrite(mem_24bit->data,20,1,output_file); // write FFD8 + FFE0 markers

    // Prepare BOX for file type 'ftyp'
    compile_box(&xt_box_header,sizeof(JPXT_FTYPE),0x66747970);
    XDP_SWAP_SHORT32(xt_ftype_payload.BR,0x6a707874); // 'jpxt'
    xt_ftype_payload.MinV=0;
    XDP_SWAP_SHORT32(xt_ftype_payload.CL,0x78727864); // 'xrxd'
    // Write 'ftype' BOX
    fwrite(&xt_box_header,1,sizeof(JPXT_SUPERBOX),output_file);
    fwrite(&xt_ftype_payload,1,sizeof(JPXT_FTYPE),output_file);

    // Legacy Data Checksum Box
    compile_box(&xt_box_header,sizeof(unsigned int),0x4C43484B); // LCHK
    fwrite(&xt_box_header,1,sizeof(JPXT_SUPERBOX),output_file);
    XDP_SWAP_SHORT32(chksum,chksum);
    fwrite(&chksum,1,sizeof(unsigned int),output_file);

    // Merging Specification Box
    XDP_SWAP_SHORT16(xt_box_merging.jpg_tag,0xFFEB);
    xt_box_merging.Le=0;
    XDP_SWAP_SHORT16(xt_box_merging.CI,0x4A50); // 'JP'
    XDP_SWAP_SHORT16(xt_box_merging.En,1);
    XDP_SWAP_SHORT32(xt_box_merging.Z,1);
    XDP_SWAP_SHORT32(xt_box_merging.LBox,sizeof(JPXT_CURV)+sizeof(xt_box_merging.LBox)+sizeof(xt_box_merging.TBox));
    XDP_SWAP_SHORT32(xt_box_merging.TBox,0x43504543);  // 'SPEC'

    // Define ldr gamma
    XDP_SWAP_SHORT32(xt_curve[0].LBox,sizeof(JPXT_CURV));
    XDP_SWAP_SHORT32(xt_curve[0].TBox,0x43555256);
    xt_curve[0].Mt=0;
    xt_curve[0].et=0;
    BIT_SET(xt_curve[0].Mt,2); // t = 4

    if (g_e==1)
        BIT_SET(xt_curve[0].et,4);

    if (srgb==0)
    {
        xt_curve[0].p1 = xt_curve[0].p3 = xt_curve[0].p4=0;
        xt_curve[0].p2=swap_float(gamma_compatibile);
    }
    else
    {
        xt_curve[0].p1=swap_float(0.04045f);
        xt_curve[0].p2=swap_float(2.4f);
        xt_curve[0].p3=swap_float(0.055f);
        xt_curve[0].p4=0;
    }
    // Define min and max values for red channel
    XDP_SWAP_SHORT32(xt_curve[1].LBox,sizeof(JPXT_CURV));
    XDP_SWAP_SHORT32(xt_curve[1].TBox,0x43555256);
    xt_curve[1].Mt=0;
    xt_curve[1].et=0;
    BIT_SET(xt_curve[1].Mt,3); // t = 8

    BIT_SET(xt_curve[1].Mt,4); // M = 9
    BIT_SET(xt_curve[1].Mt,7);
    if (g_e==1)
        BIT_SET(xt_curve[1].et,4);

    xt_curve[1].p1=swap_float(min_r);
    xt_curve[1].p2=swap_float(max_r);
    xt_curve[1].p3=swap_float(auto_gamma);
    xt_curve[1].p4=0;

    // Define min and max values for green channel
    XDP_SWAP_SHORT32(xt_curve[2].LBox,sizeof(JPXT_CURV));
    XDP_SWAP_SHORT32(xt_curve[2].TBox,0x43555256);
    xt_curve[2].Mt=0;
    xt_curve[2].et=0;
    BIT_SET(xt_curve[2].Mt,3); // t = 8
    BIT_SET(xt_curve[2].Mt,5);
    BIT_SET(xt_curve[2].Mt,7); // M = 10
    if (g_e==1)
        BIT_SET(xt_curve[2].et,4);

    xt_curve[2].p1=swap_float(min_g);
    xt_curve[2].p2=swap_float(max_g);
    xt_curve[2].p3=swap_float(auto_gamma);
    xt_curve[2].p4=0;

    // Define min and max values for blue channel
    XDP_SWAP_SHORT32(xt_curve[3].LBox,sizeof(JPXT_CURV));
    XDP_SWAP_SHORT32(xt_curve[3].TBox,0x43555256);
    xt_curve[3].Mt=0;
    xt_curve[3].et=0;
    BIT_SET(xt_curve[3].Mt,3); // t = 8
    BIT_SET(xt_curve[3].Mt,4);
    BIT_SET(xt_curve[3].Mt,5);
    BIT_SET(xt_curve[3].Mt,7); // M = 11
    if (g_e==1)
        BIT_SET(xt_curve[3].et,4);
    xt_curve[3].p1=swap_float(min_b);
    xt_curve[3].p2=swap_float(max_b);
    xt_curve[3].p3=swap_float(auto_gamma);
    xt_curve[3].p4=0;

    // Define exposure value
    XDP_SWAP_SHORT32(xt_curve[4].LBox,sizeof(JPXT_CURV));
    XDP_SWAP_SHORT32(xt_curve[4].TBox,0x43555256);
    xt_curve[4].Mt=0;
    xt_curve[4].et=0;
    BIT_SET(xt_curve[4].Mt,1); // t = 6
    BIT_SET(xt_curve[4].Mt,2);
    BIT_SET(xt_curve[4].Mt,6); // M = 4

    if (g_e==1)
        BIT_SET(xt_curve[4].et,4);
    xt_curve[4].p1 = g_ocon_p1;
    xt_curve[4].p2 = g_ocon_p2; //swap_float(1.0f/(powf(2.0f,8.0f+8.0f) -1.0f));
    if(g_use_custom_ocon==1)
        xt_curve[4].p3=g_ocon_p3;
    else
        xt_curve[4].p3=swap_float(out_exposure_value);
    xt_curve[4].p4 = g_ocon_p4;

    // Define curve for log decondig parameters for BASE
    XDP_SWAP_SHORT32(xt_curve_log[0].LBox,sizeof(JPXT_CURV));
    XDP_SWAP_SHORT32(xt_curve_log[0].TBox,0x43555256);
    xt_curve_log[0].Mt=0;
    xt_curve_log[0].et=0;
    BIT_SET(xt_curve_log[0].Mt,0);
    BIT_SET(xt_curve_log[0].Mt,1); // t = 7
    BIT_SET(xt_curve_log[0].Mt,2);
    BIT_SET(xt_curve_log[0].Mt,5); // M = 2

    if (g_e==1)
        BIT_SET(xt_curve_log[0].et,4);
    xt_curve_log[0].p1 = swap_float(1.0f);
    xt_curve_log[0].p2=swap_float(1.0f);
    xt_curve_log[0].p3=swap_float(1e-6);
    xt_curve_log[0].p4 = swap_float(0.0f);

    // Define curve for log decondig parameters for RESI
    XDP_SWAP_SHORT32(xt_curve_log[1].LBox,sizeof(JPXT_CURV));
    XDP_SWAP_SHORT32(xt_curve_log[1].TBox,0x43555256);
    xt_curve_log[1].Mt=0;
    xt_curve_log[1].et=0;
    BIT_SET(xt_curve_log[1].Mt,0);
    BIT_SET(xt_curve_log[1].Mt,1); // t = 7
    BIT_SET(xt_curve_log[1].Mt,2);
    BIT_SET(xt_curve_log[1].Mt,4); // M = 3
    BIT_SET(xt_curve_log[1].Mt,5);
    if (g_e==1)
        BIT_SET(xt_curve_log[1].et,4);
    xt_curve_log[1].p1 =swap_float(-1.0f);
    xt_curve_log[1].p2=swap_float(1.0f);
    xt_curve_log[1].p3=swap_float(1e-6);
    xt_curve_log[1].p4 =swap_float(powf(2.0f,8.0f+8.0f-1.0f)/(powf(2.0f,8+8)-1.0f)) ; // swap_float(32768.0f/655355.0f);

    if (input_xyz==1)
    {
        XDP_SWAP_SHORT32(xt_lin_trafo[0].LBox,sizeof(JPXT_FP_LIN_TRAFO));
        XDP_SWAP_SHORT32(xt_lin_trafo[0].TBox,0x46545258); // FTRX
        xt_lin_trafo[0].Mt=0;
        BIT_SET(xt_lin_trafo[0].Mt,0); // M = 7
        BIT_SET(xt_lin_trafo[0].Mt,1);
        BIT_SET(xt_lin_trafo[0].Mt,2);
        xt_lin_trafo[0].a11=swap_float(g_RGB2XYZ[0][0]);
        xt_lin_trafo[0].a12=swap_float(g_RGB2XYZ[0][1]);
        xt_lin_trafo[0].a13=swap_float(g_RGB2XYZ[0][2]);
        xt_lin_trafo[0].a21=swap_float(g_RGB2XYZ[1][0]);
        xt_lin_trafo[0].a22=swap_float(g_RGB2XYZ[1][1]);
        xt_lin_trafo[0].a23=swap_float(g_RGB2XYZ[1][2]);
        xt_lin_trafo[0].a31=swap_float(g_RGB2XYZ[2][0]);
        xt_lin_trafo[0].a32=swap_float(g_RGB2XYZ[2][1]);
        xt_lin_trafo[0].a33=swap_float(g_RGB2XYZ[2][2]);

        XDP_SWAP_SHORT32(xt_lin_trafo[1].LBox,sizeof(JPXT_FP_LIN_TRAFO));
        XDP_SWAP_SHORT32(xt_lin_trafo[1].TBox,0x46545258); // FTRX
        xt_lin_trafo[1].Mt=0;
        BIT_SET(xt_lin_trafo[1].Mt,3); // M = 8

        if (resi_identity==0)
        {
            xt_lin_trafo[1].a11=swap_float(g_RGB2XYZ[0][0]);
            xt_lin_trafo[1].a12=swap_float(g_RGB2XYZ[0][1]);
            xt_lin_trafo[1].a13=swap_float(g_RGB2XYZ[0][2]);
            xt_lin_trafo[1].a21=swap_float(g_RGB2XYZ[1][0]);
            xt_lin_trafo[1].a22=swap_float(g_RGB2XYZ[1][1]);
            xt_lin_trafo[1].a23=swap_float(g_RGB2XYZ[1][2]);
            xt_lin_trafo[1].a31=swap_float(g_RGB2XYZ[2][0]);
            xt_lin_trafo[1].a32=swap_float(g_RGB2XYZ[2][1]);
            xt_lin_trafo[1].a33=swap_float(g_RGB2XYZ[2][2]);
        }
        else
        {
            xt_lin_trafo[1].a11=swap_float(1.0f);
            xt_lin_trafo[1].a12=swap_float(0.0f);
            xt_lin_trafo[1].a13=swap_float(0.0f);
            xt_lin_trafo[1].a21=swap_float(0.0f);
            xt_lin_trafo[1].a22=swap_float(1.0f);
            xt_lin_trafo[1].a23=swap_float(0.0f);
            xt_lin_trafo[1].a31=swap_float(0.0f);
            xt_lin_trafo[1].a32=swap_float(0.0f);
            xt_lin_trafo[1].a33=swap_float(1.0f);
        }
    }


    // Write Merging Specification Super Box and needed curvs
    superbox_size=sizeof(JPXT_CURV)*5;
    //if (use_log_decoding==1)
        superbox_size+=sizeof(JPXT_CURV)*2;
    if (input_xyz==1)
    {
        if (resi_identity==0)
            superbox_size+=sizeof(JPXT_FP_LIN_TRAFO)*2;
        else
            superbox_size+=sizeof(JPXT_FP_LIN_TRAFO);
    }

    XDP_SWAP_SHORT32(xt_trafo_nlp.LBox,sizeof(JPXT_BASE_NON_LINEAR_TRAFO));
    XDP_SWAP_SHORT32(xt_trafo_nlp.TBox,0x4C505453); // LPTS
    xt_trafo_nlp.td0td1=0;
    xt_trafo_nlp.td2td3=0;


    XDP_SWAP_SHORT32(xt_trafo_ir.LBox,sizeof(JPXT_BASE_NON_LINEAR_TRAFO));
    XDP_SWAP_SHORT32(xt_trafo_ir.TBox,0x44505453); // DPTS
    xt_trafo_ir.td0td1=0;
    xt_trafo_ir.td2td3=0;
    BIT_SET(xt_trafo_ir.td0td1,1); // 1 td1=10
    BIT_SET(xt_trafo_ir.td0td1,3);
    BIT_SET(xt_trafo_ir.td0td1,4);
    BIT_SET(xt_trafo_ir.td0td1,7); // 1 td0=9
    BIT_SET(xt_trafo_ir.td2td3,4);
    BIT_SET(xt_trafo_ir.td2td3,5);
    BIT_SET(xt_trafo_ir.td2td3,7); // 1 td2=2


    XDP_SWAP_SHORT32(xt_trafo_log_base.LBox,sizeof(JPXT_BASE_NON_LINEAR_TRAFO));
    XDP_SWAP_SHORT32(xt_trafo_log_base.TBox,0x43505453); // CPTS
    xt_trafo_log_base.td0td1=0;
    xt_trafo_log_base.td2td3=0;
    BIT_SET(xt_trafo_log_base.td0td1,1); // 1 td1=2
    BIT_SET(xt_trafo_log_base.td0td1,5); // 1 td0=2
    BIT_SET(xt_trafo_log_base.td2td3,5); // 1 td2=2


    XDP_SWAP_SHORT32(xt_trafo_log_resi.LBox,sizeof(JPXT_BASE_NON_LINEAR_TRAFO));
    XDP_SWAP_SHORT32(xt_trafo_log_resi.TBox,0x52505453); // RPTS
    xt_trafo_log_resi.td0td1=0;
    xt_trafo_log_resi.td2td3=0;
    BIT_SET(xt_trafo_log_resi.td0td1,0); // 1 td1=3
    BIT_SET(xt_trafo_log_resi.td0td1,1);
    BIT_SET(xt_trafo_log_resi.td0td1,4); // 1 td0=3
    BIT_SET(xt_trafo_log_resi.td0td1,5);
    BIT_SET(xt_trafo_log_resi.td2td3,4);
    BIT_SET(xt_trafo_log_resi.td2td3,5); // 1 td2=3

    superbox_size +=sizeof(JPXT_BASE_NON_LINEAR_TRAFO)*4;

    XDP_SWAP_SHORT32(xtConversionBox.LBox,sizeof(JPXT_OUT_CONV_BOX));
    XDP_SWAP_SHORT32(xtConversionBox.TBox,0x4F434F4E); // OCON
    xtConversionBox.RbLfOcCeOl=0;
    xtConversionBox.to0to1=0;
    xtConversionBox.to2to3=0;
    BIT_SET(xtConversionBox.RbLfOcCeOl,7);
    BIT_SET(xtConversionBox.RbLfOcCeOl,0); // Ol = 1
    BIT_SET(xtConversionBox.to0to1,2); // 1 to1=4
    BIT_SET(xtConversionBox.to0to1,6); // 1 to0=4
    BIT_SET(xtConversionBox.to2to3,6); // 1 to2=4

    superbox_size += sizeof(JPXT_OUT_CONV_BOX);

    if (input_xyz==1)
    {
        XDP_SWAP_SHORT32(xt_color_trafo.LBox,sizeof(JPXT_COLOR_TRAFO));
        XDP_SWAP_SHORT32(xt_color_trafo.TBox,0x43545246); // CTRF
        xt_color_trafo.CXtCRe=0;
        BIT_SET(xt_color_trafo.CXtCRe,0); // M = 7
        BIT_SET(xt_color_trafo.CXtCRe,1);
        BIT_SET(xt_color_trafo.CXtCRe,2);
        superbox_size += sizeof(JPXT_COLOR_TRAFO);

        if (resi_identity==0)
        {
            XDP_SWAP_SHORT32(xt_resi_trafo.LBox,sizeof(JPXT_RESI_TRAFO));
            XDP_SWAP_SHORT32(xt_resi_trafo.TBox,0x44545246); // DTRF
            xt_resi_trafo.DXtDRc=0;
            BIT_SET(xt_resi_trafo.DXtDRc,3); // M = 8
            superbox_size += sizeof(JPXT_RESI_TRAFO);
        }
    }

    compile_box(&xt_box_header,superbox_size,0x53504543); // SPEC BOX
    fwrite(&xt_box_header,1,sizeof(JPXT_SUPERBOX),output_file);
    fwrite(&xt_curve[0],1,sizeof(JPXT_CURV),output_file);
    fwrite(&xt_curve[1],1,sizeof(JPXT_CURV),output_file);
    fwrite(&xt_curve[2],1,sizeof(JPXT_CURV),output_file);
    fwrite(&xt_curve[3],1,sizeof(JPXT_CURV),output_file);
    fwrite(&xt_curve[4],1,sizeof(JPXT_CURV),output_file);

    fwrite(&xt_curve_log[0],1,sizeof(JPXT_CURV),output_file);
    fwrite(&xt_curve_log[1],1,sizeof(JPXT_CURV),output_file);
    fwrite(&xt_trafo_log_base,1,sizeof(JPXT_BASE_NON_LINEAR_TRAFO),output_file);
    fwrite(&xt_trafo_log_resi,1,sizeof(JPXT_BASE_NON_LINEAR_TRAFO),output_file);

    if (input_xyz==1)
    {
        fwrite(&xt_lin_trafo[0],1,sizeof(JPXT_FP_LIN_TRAFO),output_file);
        fwrite(&xt_color_trafo,1,sizeof(JPXT_COLOR_TRAFO),output_file);
        if (resi_identity==0)
        {
            fwrite(&xt_lin_trafo[1],1,sizeof(JPXT_FP_LIN_TRAFO),output_file);
            fwrite(&xt_resi_trafo,1,sizeof(JPXT_RESI_TRAFO),output_file);
        }
    }
    fwrite(&xt_trafo_nlp,1,sizeof(JPXT_BASE_NON_LINEAR_TRAFO),output_file);
    fwrite(&xt_trafo_ir,1,sizeof(JPXT_BASE_NON_LINEAR_TRAFO),output_file);
    fwrite(&xtConversionBox,1,sizeof(JPXT_OUT_CONV_BOX),output_file);


    // Write residual Data Boxes)
    residualChkSize=65535-2-sizeof(JPXT_SUPERBOX); // Maximum RESI payload size
    nchunks=ceil((float)mem_hdr_24bit->data_size/(float)residualChkSize); // Number of chunks to write
    chksize=residualChkSize;
    off=0;
    chkIdx=0;
    for (w=0;w<nchunks;w++)
    {
        if(off+residualChkSize>mem_hdr_24bit->data_size)  // is last chunk ??
            chksize=mem_hdr_24bit->data_size-off;
        else
            chksize=residualChkSize;

        compile_box(&xtBoxResidual,chksize,0x52455349); // RESI
        XDP_SWAP_SHORT32(xtBoxResidual.LBox,mem_hdr_24bit->data_size+sizeof(xtBoxResidual.LBox)+sizeof(xtBoxResidual.TBox));
        XDP_SWAP_SHORT32(xtBoxResidual.Z,chkIdx);
        xtBoxResidual.En=0;
        fwrite(&xtBoxResidual,1,sizeof(JPXT_SUPERBOX),output_file);
        fwrite(mem_hdr_24bit->data+off,chksize,1,output_file);

        off += chksize;
        chkIdx++;
    }



    fwrite(mem_24bit->data+20,mem_24bit->data_size-20,1,output_file); // Write remaining LDR jpeg file
    fclose(output_file);
    END_TIMER();
    if (prencoding==0)
        free_image(out_24bit);
    else
        free_memory_image(out_24bit_prencoded);

    return 1;
}



